@@ -7,7 +7,7 @@
 //
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY ormain FITNESS FOR A PARTICULAR PURPOSE.  See the
+// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU General Public License for more details.
 //
 // You should have received a copy of the GNU General Public License
@@ -17,21 +17,39 @@
 //
 // Revision history (latest first):
 //
+// 09.05.2005, Stefan Becker, Aenderungen aus k9copy integriert 
+//
+// 2004/11/30: V0.97-1: Implemented injection support. This gives us *much*
+//                    more reliable results with the targeted size. The
+//                    vaporized data size now usually hits within a range of
+//                    a few thousand sectors (i.e. <~5MB) only. Don't forget
+//                    to remove the injections file with each new disk!
+//
+// 2004/11/29: V0.97: Video buffers aren't static any more. This
+//                    is needed for streams with multiple adjacent
+//                    GOPs not starting on sector boundaries.
+//
+// 2004/05/25: V0.96: Now GOPs do not necessarily need to start
+//                    on a sector boundary. This gets us rid of
+//                    "Bad sequence header code at ...".
+//
 // 2003/12/13: V0.95: No more known bugs nor any essential to-dos!
 
 // for large file support (>2GB)
-//
 #define _GNU_SOURCE
-#define VERSION		"V0.95"
+#define _FILE_OFFSET_BITS	64
+#define VERSION		"V0.98"
 
 #include <stdio.h>
 #include <errno.h>
 #include <stdarg.h>
 #include <stdlib.h>
+#include <stdint.h>
 #include <unistd.h>
 #include <string.h>
 #include <getopt.h>
 #include <pthread.h>
+#include <sys/stat.h>
 
 // DVD sector size
 #define SECT_SIZE 2048
@@ -42,7 +60,7 @@
 // write buffer size (4MB)
 #define WBUF_SIZE (0x1000*1024)
 
-// video buffer size (1MB)
+// initial video buffer size (1MB)
 #define VBUF_SIZE (1024*1024)
 
 
@@ -53,43 +71,70 @@ typedef unsigned long long UInt64;
 
 
 // a lot of global data - we really should use C++...
-UInt8      rbuf  [RBUF_SIZE];		// the PS read buffer
-UInt8      wbuf  [WBUF_SIZE];		// the PS write buffer
-UInt8      vibuf [VBUF_SIZE];		// the video ES requant input buffer
-UInt8      vobuf [VBUF_SIZE];		// the video ES requant output buffer
-UInt8     *rptr = rbuf;			// pointer to current char in read buf
-UInt8     *rhwp = rbuf;			// read buffer high water pointer
-UInt8     *wptr = wbuf;			// pointer to first unused char in wbuf
-UInt64     bytes_read;			// total PS bytes read
-UInt64     bytes_written;		// total PS bytes written
-UInt64     padding_bytes;		// total padding bytes written
-UInt64     vin_bytes;			// total unshrinked video ES bytes
-UInt64     vout_bytes;			// total shrinked video ES bytes
-int        vilen;			// current GOP's unshrinked vidES bytes
-int        volen;			// current GOP's shrinked vidES bytes
-int        total_packs;			// total no. PS packs
-int        video_packs;			// no. video packs in PS
-int        skipped_video_packs;		// skipped thereof
-int        private_1_packs;		// no. private stream 1 packs in PS
-int        skipped_private_1_packs;	// skipped thereof
-int        sequence_headers;		// no. sequence headers (== #GOPs)
-int        nav_packs;			// no. nav packs
-int        eof;				// end of file flag
-UInt8     *rqt_rptr;			// ptr to cur char in requant read buf
-UInt8     *rqt_wptr;			// ptr to first unused char in rqt wbuf
-int        rqt_rcnt;			// no. bytes in requant read buffer
-int        rqt_wcnt;			// no. bytes in requant write buffer
-UInt64     rqt_inbytes;			// total bytes fed into requantization
-UInt64     rqt_outbytes;		// total bytes melt from requantization
-float      rqt_fact;			// video ES vaporization factor (>1)
-int        spu_track_map [32];		// subpicture track# translation map
-int        audio_track_map [32];	// audio track# translation map
-int        verbose;			// level of verbosity
-int        calc_ps_vap;			// calc vaporization based on PS size
-float      vap_fact    = 1.0f;		// vaporization factor from cmd line
-const char progname [] = "vamps";	// we're sucking bytes!
+UInt8       rbuf  [RBUF_SIZE];		// the PS read buffer
+UInt8       wbuf  [WBUF_SIZE];		// the PS write buffer
+UInt8      *vibuf;			// the video ES requant input buffer
+UInt8      *vobuf;			// the video ES requant output buffer
+UInt8      *rptr = rbuf;		// pointer to current char in read buf
+UInt8      *rhwp = rbuf;		// read buffer high water pointer
+UInt8      *wptr = wbuf;		// pointer to first unused char in wbuf
+UInt64      bytes_read;			// total PS bytes read
+UInt64      bytes_written;		// total PS bytes written
+UInt64      padding_bytes;		// total padding bytes written
+UInt64      vin_bytes;			// total unshrinked video ES bytes
+UInt64      vout_bytes;			// total shrinked video ES bytes
+int         vbuf_size = VBUF_SIZE;	// the video ES requant buffers' size
+int         vilen;			// current GOP's unshrinked vidES bytes
+int         volen;			// current GOP's shrinked vidES bytes
+int         total_packs;		// total no. PS packs
+int         video_packs;		// no. video packs in PS
+int         skipped_video_packs;	// skipped thereof
+int         private_1_packs;		// no. private stream 1 packs in PS
+int         skipped_private_1_packs;	// skipped thereof
+int         sequence_headers;		// no. sequence headers (== #GOPs)
+int         nav_packs;			// no. nav packs
+int         eof;			// end of file flag
+UInt8      *rqt_rptr;			// ptr to cur char in requant read buf
+UInt8      *rqt_wptr;			// ptr to first unused char in rqt wbuf
+int         rqt_rcnt;			// no. bytes in requant read buffer
+int         rqt_wcnt;			// no. bytes in requant write buffer
+UInt64      rqt_inbytes;		// total bytes fed into requantization
+UInt64      rqt_outbytes;		// total bytes melt from requantization
+UInt64      rqt_visize;			// calc. total unshrinked vidES bytes
+float       rqt_fact;			// video ES vaporization factor (>1)
+int         spu_track_map [32];		// subpicture track# translation map
+int         audio_track_map [32];	// audio track# translation map
+int         verbose;			// level of verbosity
+int         calc_ps_vap;		// calc vaporization based on PS size
+float       vap_fact    = 1.0f;		// vaporization factor from cmd line
+const char *injections_file;		// where to inject internal status from
+const char  progname [] = "vamps";	// we're sucking bytes!
 
-float      fLastFaktor=0;               // Zuletzt verwendeter Faktor fürs Eindampfen
+// injections file stuff
+#define inject_int(v)		{ sizeof (v), &v, "%d"   }
+#define inject_uint64(v)	{ sizeof (v), &v, "%llu" }
+#define N_INJECTIONS		(sizeof (injections) / sizeof (*injections))
+
+struct
+{
+  int         s;	// variable size
+  const void *p;	// variable pointer
+  const char *f;	// printf/scanf format
+} injections [] =
+{
+  inject_uint64 (bytes_read),
+  inject_uint64 (bytes_written),
+  inject_uint64 (padding_bytes),
+  inject_uint64 (vin_bytes),
+  inject_uint64 (vout_bytes),
+  inject_int    (total_packs),
+  inject_int    (video_packs),
+  inject_int    (skipped_video_packs),
+  inject_int    (private_1_packs),
+  inject_int    (skipped_private_1_packs),
+  inject_int    (sequence_headers),
+  inject_int    (nav_packs)
+};
 
 // stuff for inter-thread com
 pthread_cond_t  condr = PTHREAD_COND_INITIALIZER;
@@ -101,6 +146,8 @@ pthread_mutex_t mutw  = PTHREAD_MUTEX_INITIALIZER;
 // prototypes
 void scan_track_map (int *, int);
 void vaporize (void);
+void read_injections (const char *);
+void write_injections (const char *);
 void usage (void);
 void fatal (char *, ...);
 
@@ -121,12 +168,13 @@ main (int argc, char *const argv [])
     { "evaporate",      1, NULL, 'e' },
     { "ps-evaporate",   1, NULL, 'E' },
     { "verbose",        0, NULL, 'v' },
+    { "inject",         1, NULL, 'i' },
     { NULL,             0, NULL, 0   }
   };
 
   opterr = 1;
 
-  while ((c = getopt_long_only (argc, argv, "a:s:e:E:v",
+  while ((c = getopt_long_only (argc, argv, "a:s:e:E:vi:",
 				long_options, NULL)) >= 0)
   {
     switch (c) {
@@ -155,7 +203,6 @@ main (int argc, char *const argv [])
       case 'E':
 	calc_ps_vap = 1;
 	vap_fact    = strtof (optarg, &s);
-        fLastFaktor = vap_fact;
 
 	if (*s)
 	{
@@ -171,12 +218,27 @@ main (int argc, char *const argv [])
 	verbose++;
 	break;
 
+      case 'i':
+	injections_file = optarg;
+        break;
+
       default:
 	putc ('\n', stderr);
 	usage ();
     }
   }
 
+  // read injections file - leaves variables untouched if file not present
+  if (injections_file)
+    read_injections (injections_file);
+
+  // allocate video buffers
+  vibuf = malloc (vbuf_size);
+  vobuf = malloc (vbuf_size);
+
+  if (vibuf == NULL || vobuf == NULL)
+    fatal ("Allocation of video buffers failed: %s", strerror (errno));
+
   // create requantization thread
   if (pthread_create (&thread, NULL, requant_thread, NULL))
       fatal ("pthread_create: %s", strerror (errno));
@@ -184,6 +246,10 @@ main (int argc, char *const argv [])
   // actually do vaporization
   vaporize ();
 
+  // save internal status for adjacent run
+  if (injections_file)
+    write_injections (injections_file);
+
   // print this run's statistics if wanted
   if (verbose >= 1)
   {
@@ -364,10 +430,10 @@ check_pack (UInt8 *ptr)
 
 
 // video packet consistency checking
-static inline void
+static inline int
 check_video_packet (UInt8 *ptr)
 {
-  int    vid_packet_length, pad_packet_length;
+  int    vid_packet_length, pad_packet_length, rc = 0;
   UInt32 vid_packet_start_code, pad_packet_start_code, sequence_header_code;
 
   vid_packet_start_code  = (UInt32) (ptr [0]) << 24;
@@ -375,30 +441,38 @@ check_video_packet (UInt8 *ptr)
   vid_packet_start_code |= (UInt32) (ptr [2]) <<  8;
   vid_packet_start_code |= (UInt32) (ptr [3]);
 
-//  if (vid_packet_start_code != 0x000001e0)
-//    fatal ("Bad video packet start code at %llu: %08lx",
-//	   rtell (ptr), vid_packet_start_code);
+  if (vid_packet_start_code != 0x000001e0)
+    fatal ("Bad video packet start code at %llu: %08lx",
+	   rtell (ptr), vid_packet_start_code);
 
   vid_packet_length  = ptr [4] << 8;
   vid_packet_length |= ptr [5];
   vid_packet_length += 6;
 
-//  if ((ptr [6] & 0xc0) != 0x80)
- //   fatal ("Not an MPEG2 video packet at %llu", rtell (ptr));
+  if ((ptr [6] & 0xc0) != 0x80)
+    fatal ("Not an MPEG2 video packet at %llu", rtell (ptr));
 
   if (ptr [7])
   {
-//    if ((ptr [7] & 0xc0) != 0xc0)
-//      fatal ("First video packet in sequence starting at %llu "
-//	     "misses PTS or DTS, flags=%02x", rtell (ptr), ptr [7]);
+    if ((ptr [7] & 0xc0) != 0xc0)
+      fatal ("First video packet in sequence starting at %llu "
+	     "misses PTS or DTS, flags=%02x", rtell (ptr), ptr [7]);
 
     sequence_header_code  = (UInt32) (ptr [6 + 3 + ptr [8] + 0]) << 24;
     sequence_header_code |= (UInt32) (ptr [6 + 3 + ptr [8] + 1]) << 16;
     sequence_header_code |= (UInt32) (ptr [6 + 3 + ptr [8] + 2]) <<  8;
     sequence_header_code |= (UInt32) (ptr [6 + 3 + ptr [8] + 3]);
-//    if (sequence_header_code != 0x000001b3)
-//      fatal ("Bad sequence header code at %llu: %08lx",
-//	     rtell (ptr + 6 + 3 + ptr [8]), sequence_header_code);
+
+    if (sequence_header_code == 0x000001b3)
+    {
+      rc = 1;
+    }
+    else
+    {
+      //fprintf (stderr, "Start of GOP at %llu not on sector boundary\n",
+      //         rtell (ptr + 6 + 3 + ptr [8]));
+      sequence_headers++;
+    }
   }
 
   pad_packet_length = 0;
@@ -427,6 +501,8 @@ check_video_packet (UInt8 *ptr)
   if (14 + vid_packet_length + pad_packet_length != SECT_SIZE)
     fatal ("Bad video packet length at %llu: %d",
 	   rtell (ptr), vid_packet_length);
+
+  return rc;
 }
 
 
@@ -506,20 +582,21 @@ new_private_1_type (UInt8 *ptr)
     // DTS audio
     abase = 0x88;
   }
-  else if (type >= 0xa0 && type <= 0xbf)
+  else if (type >= 0xa0 && type <= 0xa7)
   {
     // LPCM audio
     abase = 0xa0;
   }
-  else if (type >= 0xc0 && type <= 0xdf)
+  else if (type >= 0xc0 && type <= 0xc7)
   {
     // MPEG audio
     abase = 0xc0;
   }
   else
   {
-    fatal ("Unknown private stream 1 type at %llu: %02x", rtell (ptr), type);
-    abase = 0;
+    printf ("Unknown private stream 1 type at %llu: %02x\n", rtell (ptr), type);
+    printf ("no break, dont copy\n");
+    return 0;
   }
 
   track = audio_track_map [type - abase];
@@ -539,7 +616,48 @@ copy_private_1 (UInt8 *ptr)
 
   if (type)
   {
-    ptr [6 + 3 + ptr [8]] = type;
+    if ((ptr[3]>=0xc0) && (ptr[3]<=0xdf)) {
+       ptr[3]= type;  
+    } else {
+       ptr [6 + 3 + ptr [8]] = type;
+    }
+    copy (SECT_SIZE);
+
+    return;
+  }
+
+  skip (SECT_SIZE);
+}
+
+
+// translate ID of MPEG audio packet
+// according to the audio track translation map
+// returns new ID (e.g. 0xc0 for first MPEG audio
+// track in cmd line) or zero if track is not to be copied
+static inline int
+new_mpeg_audio_id (int id)
+{
+  int track;
+
+  track = audio_track_map [id - 0xc0];
+
+  return track ? track - 1 + 0xc0 : 0;
+}
+
+
+// selectivly copy MPEG audio packs
+// patches ID to reflect new track mapping unless user opted to preserve them
+static inline void
+copy_mpeg_audio (UInt8 *ptr)
+{
+  int id;
+
+  id = new_mpeg_audio_id (ptr [3]);
+
+  if (id)
+  {
+    ptr [3] = id;
+
     copy (SECT_SIZE);
 
     return;
@@ -573,11 +691,8 @@ vap_leader (void)
     {
       case 0xe0:
 	// video
-	check_video_packet (ptr);
-
-	if ((ptr [7] & 0xc0) == 0xc0)
-	  // DTS/PTS present - this is a sequence header
-	  // (guarantied by check_video_packet())
+	if (check_video_packet (ptr))
+	  // sequence header
 	  return;
 
 	copy (SECT_SIZE);
@@ -588,8 +703,16 @@ vap_leader (void)
 	copy_private_1 (ptr);
 	break;
 
-      case 0xbe: // skip padding EMS
-	skip(SECT_SIZE);
+      case 0xc0:
+      case 0xc1:
+      case 0xc2:
+      case 0xc3:
+      case 0xc4:
+      case 0xc5:
+      case 0xc6:
+      case 0xc7:
+	// MPEG audio
+	copy_mpeg_audio (ptr);
 	break;
 
       case 0xbb:
@@ -598,15 +721,10 @@ vap_leader (void)
 	break;
 
       default:
-	check_video_packet (ptr);
+         // Ignore, i. E. 0xbe with Button
+ 	 skip (SECT_SIZE);
+	 break;
 
-	if ((ptr [7] & 0xc0) == 0xc0)
-	  // DTS/PTS present - this is a sequence header
-	  // (guarantied by check_video_packet())
-	  return;
-
-	copy (SECT_SIZE);
-	break;
     }
 
     if (wptr == wbuf + WBUF_SIZE)
@@ -638,6 +756,11 @@ vap_trailer (int length)
       // private 1: audio/subpicture
       copy_private_1 (ptr);
     }
+    else if (id >= 0xc0 && id <= 0xc7)
+    {
+      // MPEG audio
+      copy_mpeg_audio (ptr);
+    }
     else
     {
       copy (SECT_SIZE);
@@ -662,7 +785,7 @@ static inline int
 vap_phase1 (void)
 {
   UInt8 *ptr, *viptr = vibuf;
-  int    seq_length, id, data_length, opt_length;
+  int    seq_length, id, data_length, opt_length, seqhdr;
 
   for (seq_length = 0;
        !lock (seq_length + SECT_SIZE); seq_length += SECT_SIZE)
@@ -683,13 +806,13 @@ vap_phase1 (void)
     {
       case 0xe0:
 	// video
-	check_video_packet (ptr);
+	seqhdr = check_video_packet (ptr);
 
 	if (seq_length)
 	{
 	  video_packs++;
 
-	  if ((ptr [7] & 0xc0) == 0xc0)
+	  if (seqhdr)
 	  {
 	    sequence_headers++;
 	    vilen = viptr - vibuf;
@@ -704,8 +827,22 @@ vap_phase1 (void)
 	opt_length   = 3 + ptr [8];
 	data_length -= opt_length;
 
-	if ((viptr - vibuf) + data_length > VBUF_SIZE - 3)
-	  fatal ("Video input buffer overflow");
+	if ((viptr - vibuf) + data_length > vbuf_size - 3)
+	{
+	  // reallocate video buffers
+	  int i = viptr - vibuf;
+
+	  // grow by another VBUF_SIZE bytes
+	  vbuf_size += VBUF_SIZE;
+	  vibuf      = realloc (vibuf, vbuf_size);
+	  vobuf      = realloc (vobuf, vbuf_size);
+
+	  if (vibuf == NULL || vobuf == NULL)
+	    fatal ("Reallocation of video buffers failed: %s",
+		   strerror (errno));
+
+	  viptr = vibuf + i;
+	}
 
 	//fprintf (stderr, "data_length=%d\n", data_length);
 	memcpy (viptr, ptr + 6 + opt_length, data_length);
@@ -721,15 +858,29 @@ vap_phase1 (void)
 
 	break;
 
+      case 0xc0:
+      case 0xc1:
+      case 0xc2:
+      case 0xc3:
+      case 0xc4:
+      case 0xc5:
+      case 0xc6:
+      case 0xc7:
+	// MPEG audio
+	private_1_packs++;
+
+	if (!new_mpeg_audio_id (id))
+	  skipped_private_1_packs++;
+
+	break;
+
       case 0xbb:
 	// system header/private 2: PCI/DSI
 	nav_packs++;
 	break;
 
-      case 0xbe: // skip padding stream EMS
-	break;
-
       default:
+        // Ignore, i. E. 0xbe with Button
 	break;
     }
   }
@@ -846,17 +997,26 @@ vap_phase2 (int seq_length)
 	copy_private_1 (ptr);
 	break;
 
+      case 0xc0:
+      case 0xc1:
+      case 0xc2:
+      case 0xc3:
+      case 0xc4:
+      case 0xc5:
+      case 0xc6:
+      case 0xc7:
+	// MPEG audio
+	copy_mpeg_audio (ptr);
+	break;
+
       case 0xbb:
 	// system header/private 2: PCI/DSI
 	copy (SECT_SIZE);
 	break;
 
-      case 0xbe: // skip padding stream EMS
-	skip(SECT_SIZE);
-	break;
-
       default:
-	skip(SECT_SIZE);
+        // Ignore, i. E. 0xbe with Button
+        skip (SECT_SIZE);
 	break;
     }
 
@@ -908,7 +1068,7 @@ vaporize (void)
     if (calc_ps_vap && vap_fact > 1.0f)
     {
       // forecast video ES vaporization factor
-      // the basic formulars looks like:
+      // the basic formulars look like:
       // vap_fact  = total_packs/(restpacks+vop)
       // restpacks = total_packs-(video_packs+skipped_private_1_packs)
       // fact      = (video_packs*net-(gops*net/2+10))/(vop*net-(gops*net/2+10))
@@ -929,9 +1089,6 @@ vaporize (void)
       if (fact < 1.0f)
 	fact = 1.0f;
 
-      // Faktor in Variable speichern
-      fLastFaktor = fact;
-
       if (verbose >= 2)
 	fprintf (stderr, "Info: Target video ES vaporization factor: %.3f\n",
 		 fact);
@@ -953,7 +1110,7 @@ vaporize (void)
 
     vout_bytes += volen;
 
-    // do phase 1 of vaporization
+    // do phase 2 of vaporization
     vap_phase2 (seq_length);
 
     //fprintf (stderr,
@@ -964,6 +1121,71 @@ vaporize (void)
 }
 
 
+// inject status from previous run into internal variables
+void
+read_injections (const char *filename)
+{
+  int   i;
+  FILE *fp;
+
+  // lack of injections file is Ok
+  // we will create one at end of this run
+  if ((fp = fopen (filename, "r")) == NULL)
+    return;
+
+  for (i = 0; i < N_INJECTIONS; i++)
+  {
+    if (fscanf (fp, injections [i].f, injections [i].p) != 1)
+      fatal ("Bad format of injections file: %s", filename);
+  }
+
+  fclose (fp);
+}
+
+
+// save status of internal variables to disk
+void
+write_injections (const char *filename)
+{
+  int   i;
+  FILE *fp;
+
+  if ((fp = fopen (filename, "w")) == NULL)
+    fatal ("Failed creating injections file: %s: %s",
+	   filename, strerror (errno));
+
+  for (i = 0; i < N_INJECTIONS; i++)
+  {
+    switch (injections [i].s)
+    {
+//      case 1:
+//	fprintf (fp, injections [i].f, *((UInt8 *) injections [i].p));
+//	break;
+//
+//      case 2:
+//	fprintf (fp, injections [i].f, *((UInt16 *) injections [i].p));
+//	break;
+//
+      case 4:
+	fprintf (fp, injections [i].f, *((UInt32 *) injections [i].p));
+	break;
+
+      case 8:
+	fprintf (fp, injections [i].f, *((UInt64 *) injections [i].p));
+	break;
+
+      default:
+	abort ();
+    }
+
+    putc ('\n', fp);
+  }
+
+  if (fclose (fp))
+    fatal ("Failed writing injections file: %s: %s");
+}
+
+
 // let's hope this helps
 void
 usage (void)
@@ -972,11 +1194,12 @@ usage (void)
 "vamps " VERSION "\n"
 "\n"
 "Usage: vamps [--evaporate|-e factor] [--ps-evaporate|-E factor]\n"
-"             [--audio|-a a-track,a-track,...]\n"
-"             [--subpictures|-s s-track,s-track,...] [--verbose|-v]\n"
+"             [--audio|-a a-stream,a-stream,...]\n"
+"             [--subpictures|-s s-stream,s-stream,...] [--verbose|-v]\n"
+"             [--inject|-i injections-file]\n"
 "\n"
 "Vamps evaporates DVD compliant MPEG2 program streams by\n"
-"selectively copying audio and subpicture tracks and by re-quantizing\n"
+"selectively copying audio and subpicture streams and by re-quantizing\n"
 "the embedded elementary video stream. The shrink factor may be either\n"
 "specified for the video ES only (-e) or for the full PS (-E).\n",
   stderr);
